#include <algorithm>
#include <fstream>
#include <iostream>

#include <misc/utils.h>

#include "gcsa.h"


namespace CSA
{

const std::string GCSA::ALPHABET = "ACGTZ";

//--------------------------------------------------------------------------

GCSA::GCSA(const std::string& base_name) :
  ok(false),
  first_children(0), outgoing(0),
  sampled_positions(0), samples(0)
{
  for(usint i = 0; i < ALPHABET_SIZE; i++) { this->array[i] = 0; }

  std::string index_name = base_name + GCSA_EXTENSION;
  std::ifstream input(index_name.c_str(), std::ios_base::binary);
  if(!input)
  {
    std::cerr << "Error opening input file (" << index_name << ")!" << std::endl;
    return;
  }

  input.read((char*)(this->counts), ALPHABET_SIZE * sizeof(usint));
  this->fillArrayC();

  for(usint i = 0; i < ALPHABET_SIZE; i++)
  {
    this->array[i] = new DeltaVector(input);
  }
  this->first_children = new RLEVector(input);
  this->outgoing = new RLEVector(input);

  this->sampled_positions = new DeltaVector(input);
  usint sample_bits = 0;
  input.read((char*)&sample_bits, sizeof(usint));
  this->samples = new ReadBuffer(input, this->sampled_positions->getNumberOfItems(), sample_bits);

  this->ok = true;
  input.close();
}

GCSA::GCSA(PathGraph& graph, bool print) :
  ok(false),
  first_children(0), outgoing(0),
  sampled_positions(0), samples(0)
{
  if(graph.status != PathGraph::sorted) { return; }

  DeltaEncoder* array_encoders[ALPHABET_SIZE];
  RLEEncoder children(CHILDREN_BLOCK_SIZE), outedges(OUTGOING_BLOCK_SIZE);

  for(usint i = 0; i < ALPHABET_SIZE; i++)
  {
    this->array[i] = 0;
    this->counts[i] = 0;
    array_encoders[i] = new DeltaEncoder(ARRAY_BLOCK_SIZE);
  }
  if(print) { std::cout << "Generating edges... "; std::cout.flush(); }
  if(!graph.generateEdges()) { return; }
  if(print) { std::cout << graph.edge_count << " edges." << std::endl; }


  // For each node, write max(indegree, outdegree) bits to BWT, F, and M.
  usint offset = 0;
  for(std::vector<PathNode>::iterator node = graph.nodes.begin(); node != graph.nodes.end(); ++node)
  {
    // Write BWT.
    usint n = 0;
    for(usint i = 0; i < ALPHABET_SIZE; i++)
    {
      if(node->hasPred(i))
      {
        this->counts[i]++;
        array_encoders[i]->setBit(offset + n); n++;
      }
    }
    n = std::max(n, (*node).outdegree());

    children.addBit(offset);  // Write F.
    // Write M
    if(node == graph.nodes.begin()) { outedges.addBit(offset); }
    else                            { outedges.addRun(offset, (*node).outdegree()); }
    offset += n;
  }
  this->fillArrayC();

  for(usint i = 0; i < ALPHABET_SIZE; i++)
  {
    this->array[i] = new DeltaVector(*(array_encoders[i]), offset);
  }
  children.flush();
  this->first_children = new RLEVector(children, offset);
  outedges.flush();
  this->outgoing = new RLEVector(outedges, offset);


  // Sample the graph for locate().
  if(print) { std::cout << "Sampling the graph... "; std::cout.flush();  }
  std::vector<pair_type>* sample_pairs = graph.getSamples(SAMPLE_RATE);
  usint max_sample = 0;
  for(usint i = 0; i < sample_pairs->size(); i++)
  {
    if(sample_pairs->at(i).second > max_sample) { max_sample = sample_pairs->at(i).second; }
  }

  DeltaEncoder sample_encoder(SAMPLE_BLOCK_SIZE);
  WriteBuffer sample_values(sample_pairs->size(), length(max_sample));
  RLEVector::Iterator iter(*(this->first_children));
  for(usint i = 0; i < sample_pairs->size(); i++)
  {
    sample_encoder.addBit(iter.select(sample_pairs->at(i).first));
    sample_values.writeItem(sample_pairs->at(i).second);
  }
  sample_encoder.flush();

  this->sampled_positions = new DeltaVector(sample_encoder, this->getSize());
  this->samples = sample_values.getReadBuffer();
  delete sample_pairs; sample_pairs = 0;
  if(print)
  {
    std::cout << this->sampled_positions->getNumberOfItems() << " samples." << std::endl;
    std::cout << std::endl;
  }


  if(print)
  {
    std::cout << "Nodes:           " << graph.node_count << std::endl;
    std::cout << "Edges:           " << graph.edge_count << std::endl;
    std::cout << "BWT length:      " << offset << std::endl;
    std::cout << "Samples:         " << this->sampled_positions->getNumberOfItems() << std::endl;
    std::cout << std::endl;

    this->reportSize(true);
  }
  this->ok = true;
}

GCSA::~GCSA()
{
  for(usint i = 0; i < ALPHABET_SIZE; i++)
  {
    delete this->array[i]; this->array[i] = 0;
  }
  delete this->first_children; this->first_children = 0;
  delete this->outgoing; this->outgoing = 0;
  delete this->sampled_positions; this->sampled_positions = 0;
  delete this->samples; this->samples = 0;
}

void
GCSA::writeTo(const std::string& base_name) const
{
  std::string index_name = base_name + GCSA_EXTENSION;
  std::ofstream output(index_name.c_str(), std::ios_base::binary);
  if(!output)
  {
    std::cerr << "Error opening output file (" << index_name << ")!" << std::endl;
    return;
  }

  output.write((char*)(this->counts), ALPHABET_SIZE * sizeof(usint));
  for(usint i = 0; i < ALPHABET_SIZE; i++)
  {
    this->array[i]->writeTo(output);
  }
  this->first_children->writeTo(output);
  this->outgoing->writeTo(output);

  this->sampled_positions->writeTo(output);
  usint sample_bits = this->samples->getItemSize();
  output.write((char*)&sample_bits, sizeof(usint));
  this->samples->writeBuffer(output);

  output.close();
}

bool
GCSA::isOk() const
{
  return this->ok;
}

//--------------------------------------------------------------------------

usint
GCSA::reportSize(bool print) const
{
  usint array_size = 0;
  for(usint i = 0; i < ALPHABET_SIZE; i++)
  {
    array_size += this->array[i]->reportSize();
  }

  usint children = this->first_children->reportSize();
  usint outedges = this->outgoing->reportSize();

  usint sa_samples = this->sampled_positions->reportSize() + this->samples->reportSize();

  usint bytes = sizeof(*this) + array_size + children + outedges + sa_samples;
  if(print)
  {
    std::cout << "Array:           " << (array_size / (double)MEGABYTE) << " MB" << std::endl;
    std::cout << "First children:  " << (children / (double)MEGABYTE) << " MB" << std::endl;
    std::cout << "Outgoing edges:  " << (outedges / (double)MEGABYTE) << " MB" << std::endl;
    std::cout << "Samples:         " << (sa_samples / (double)MEGABYTE) << " MB" << std::endl;
    std::cout << "Total size:      " << (bytes / (double)MEGABYTE) << " MB" << std::endl;
    std::cout << std::endl;
  }

  return bytes;
}

usint
GCSA::getSize() const
{
  return this->first_children->getSize();
}

//--------------------------------------------------------------------------

pair_type
GCSA::find(const std::string& pattern) const
{
  if(pattern.length() == 0) { return this->getSARange(); }

  sint pos = pattern.length() - 1;
  pair_type range = this->getCharRange(pattern[pos]);
  if(isEmpty(range)) { return range; }

  for(pos--; pos >= 0; pos--)
  {
    range = this->LF(range, pattern[pos]);
    if(isEmpty(range)) { return range; }
  }
  return range;
}

//--------------------------------------------------------------------------

std::vector<usint>*
GCSA::locate(pair_type range) const
{
  std::vector<usint>* vec = new std::vector<usint>;

  this->locate(range, *vec);
  removeDuplicates(vec, false);

  return vec;
}

std::vector<usint>*
GCSA::locate(std::vector<pair_type>& ranges) const
{
  std::vector<usint>* vec = new std::vector<usint>;

  for(std::vector<pair_type>::iterator iter = ranges.begin(); iter != ranges.end(); ++iter)
  {
    this->locate(*iter, *vec);
  }
  removeDuplicates(vec, false);

  return vec;
}

//--------------------------------------------------------------------------

pair_type
GCSA::getSARange() const
{
  return pair_type(0, this->getSize() - 1);
}

pair_type
GCSA::getBWTRange() const
{
  return pair_type(0, this->getSize() - 1);
}

pair_type
GCSA::getCharRange(usint c) const
{
  if(c == 0) { return pair_type(0, 0); }

  size_t temp = ALPHABET.find(c);
  if(temp == std::string::npos) { return EMPTY_PAIR; }

  // Edge range.
  pair_type range(this->cumulative[temp], this->cumulative[temp + 1] - 1);

  return this->convertToNodeRange(range);
}

void
GCSA::convertToBWTRange(pair_type& sa_range) const
{
}

void
GCSA::convertToSARange(pair_type& bwt_range) const
{
}

void
GCSA::convertToSARange(std::vector<pair_type>& bwt_ranges) const
{
}

pair_type
GCSA::LF(pair_type range, usint c) const
{
  size_t temp = ALPHABET.find(c);
  if(temp == std::string::npos || this->counts[temp] == 0) { return EMPTY_PAIR; }

  // Follow edges backward using BWT.
  DeltaVector::Iterator array_iter(*(this->array[temp]));
  range.first = this->cumulative[temp] + array_iter.rank(range.first, true) - 1;
  range.second = this->cumulative[temp] + array_iter.rank(range.second) - 1;
  if(isEmpty(range)) { return EMPTY_PAIR; }

  return this->convertToNodeRange(range);
}

std::vector<usint>*
GCSA::locateRange(pair_type range) const
{
  return this->locate(range);
}

std::vector<usint>*
GCSA::locateRanges(std::vector<pair_type>& ranges) const
{
  return this->locate(ranges);
}

//--------------------------------------------------------------------------

bool
GCSA::exists(usint c) const
{
  size_t temp = ALPHABET.find(c);
  if(temp == std::string::npos) { return false; }

  return (this->counts[temp] > 0);
}

usint
GCSA::getCharacter(usint index) const
{
  for(usint i = 0; i < ALPHABET_SIZE; i++)
  {
    if(this->cumulative[i + 1] > index) { return i; }
  }

  return ALPHABET_SIZE;
}

pair_type
GCSA::convertToNodeRange(pair_type edge_range) const
{
  // Convert edge range to BWT range using M.
  RLEVector::Iterator outgoing_iter(*(this->outgoing));
  edge_range.first = outgoing_iter.select(edge_range.first);
  edge_range.second = outgoing_iter.select(edge_range.second);

  // Convert BWT range to node range using F.
  RLEVector::Iterator children_iter(*(this->first_children));
  edge_range.first = children_iter.valueBefore(edge_range.first).first;
  edge_range.second = children_iter.valueAfter(edge_range.second + 1).first - 1;

  // These at least work.
//   edge_range.first = children_iter.select(children_iter.rank(edge_range.first) - 1);
//   edge_range.second = children_iter.select(children_iter.rank(edge_range.second)) - 1;

  return edge_range;
}

pair_type
GCSA::Psi(pair_type range) const
{
  if(range.first == 0) { return EMPTY_PAIR; } // Final node.

  // Find the rank of the first outgoing edge using M.
  RLEVector::Iterator outgoing_iter(*(this->outgoing));
  usint temp = outgoing_iter.rank(range.first - 1);
  usint c = this->getCharacter(temp);

  // Find the corresponding incoming edge using BWT.
  DeltaVector::Iterator array_iter(*(this->array[c]));
  temp = array_iter.select(temp - this->cumulative[c]);

  // Find the corresponding node using F.
  RLEVector::Iterator children_iter(*(this->first_children));
//   range.first = children_iter.select(children_iter.rank(temp) - 1);
  range.first = children_iter.valueBefore(temp).first;
  range.second = children_iter.selectNext() - 1;

  return range;
}

//--------------------------------------------------------------------------

void
GCSA::locate(pair_type range, std::vector<usint>& vec) const
{
  range.second = std::min(range.second, this->getSize() - 1);
  if(isEmpty(range)) { return; }

  RLEVector::Iterator iter(*(this->first_children));
  DeltaVector::Iterator sample_iter(*(this->sampled_positions));
  for(usint pos = iter.valueAfter(range.first).first; pos <= range.second; pos = iter.selectNext())
  {
    pair_type rng(pos, pos);
    usint steps = 0;
    pair_type sample = sample_iter.valueAfter(rng.first);
    while(sample.first != rng.first)
    {
      rng = this->Psi(rng); steps++;
      sample = sample_iter.valueAfter(rng.first);
    }
    vec.push_back(this->samples->readItem(sample.second) - steps);
  }
}

//--------------------------------------------------------------------------

void
GCSA::fillArrayC()
{
  this->cumulative[0] = 1;

  for(usint i = 0; i < ALPHABET_SIZE; i++)
  {
    this->cumulative[i + 1] = this->cumulative[i] + this->counts[i];
  }
}

//--------------------------------------------------------------------------

} // namespace CSA
