#ifndef GRAPH_H
#define GRAPH_H


#include <fstream>
#include <iostream>
#include <vector>


#include <misc/definitions.h>


namespace CSA
{

//--------------------------------------------------------------------------

struct GraphNode
{
  uint label;
  uint value;  // Used for sampling in GCSA.

  GraphNode() {}
  GraphNode(uint _label, uint _value) : label(_label), value(_value) {}
};

struct GraphEdge
{
  uint from, to;

  GraphEdge() {}
  GraphEdge(uint _from, uint _to) : from(_from), to(_to) {}
};

//--------------------------------------------------------------------------

struct PathNode
{
  uint      from, to;
  pair_type key;

  void setSorted()      { this->to = this->from; }
  bool isSorted() const { return (this->to == this->from); }

  // These can be used only when PathGraph.status == ready.
  uint  value() const         { return this->to; }
  usint outdegree() const     { return this->key.second; }
  bool  hasPred(uint i) const { return (this->key.first & (((usint)1) << i)); }
};

struct PathEdge
{
  usint from, to;

  PathEdge(usint _from, usint _to) : from(_from), to(_to) {}
};

//--------------------------------------------------------------------------

class Graph
{
  public:
    Graph(std::vector<GraphNode>& node_vector, std::vector<GraphEdge>& edge_vector);
    Graph(const std::string& base_name);
    ~Graph();

    void write(const std::string& base_name);

    void printInfo();

    void sortEdges(bool from);
    pair_type getEdges(uint node, bool from);  // Use sortEdges() first.

    GraphNode* nodes;
    GraphEdge* edges;
    uint       node_count, edge_count;
    bool       ok;  // Also tells that edges are sorted by from.
};

//--------------------------------------------------------------------------

class PathGraph
{
  public:
    PathGraph(Graph& _parent);      // Creates a generation 1 PathGraph.
    PathGraph(PathGraph& previous); // Creates a next generation PathGraph. Invalidates previous.
    ~PathGraph();

    void printInfo();

    typedef std::vector<PathNode> nvector;
    typedef std::vector<PathEdge> evector;

    // status must be sorted. Calling this invalidates parent and
    // sets status to ready.
    // Writes outdegree to PathNode.key.second, value to PathNode.to, and
    // predecessor labels to PathNode.key.first.
    bool generateEdges();

    // Returns SA samples as (index, value). status must be ready.
    // User must delete the samples. Call generateEdges() first.
    std::vector<pair_type>* getSamples(usint sample_rate);

    nvector nodes;
    evector edges;
    usint   node_count, edge_count, ranks;

    Graph&  parent;
    usint   generation; // Sorted by paths of length 2^generation.

    enum status_t { error, ok, sorted, ready } status;

  private:
    void      createPathNode(uint from, uint to);
    void      createPathNode(const PathNode& left, const PathNode& right);
    void      sort();
    pair_type nextMaximalSet(pair_type node_range);

    void      sortByKey();
    void      sortEdges();
    void      sortByFrom();
    void      sortEdgesByFrom();
    pair_type getNodesFrom(uint node);  // Use sortByFrom() first.
    pair_type getEdgesFrom(usint node);  // Use sortEdgesByFrom() first.
};

//--------------------------------------------------------------------------


} // namespace CSA


#endif // GRAPH_H

