#include <cstdlib>
#include <ctime>
#include <iostream>
#include <vector>

#include <rlcsa.h>

#include "bwasearch.h"
#include "parameter_handler.h"
#include "pattern_classifier.h"


using namespace CSA;


int main(int argc, char** argv)
{
  std::cout << "RLCSA test" << std::endl;
  std::cout << std::endl;

  ParameterHandler handler(argc, argv, true, "Usage: rlcsa_test [options] base_name [patterns]");
  if(!(handler.ok)) { handler.printUsage(); return 1; }
  handler.printOptions();

  const RLCSA rlcsa(handler.index_name);
  if(!rlcsa.isOk()) { return 2; }
  rlcsa.reportSize(true);
  if(handler.patterns_name == 0) { return 0; }
  BWASearch<RLCSA> bwasearch(rlcsa);

  std::ifstream patterns(handler.patterns_name, std::ios_base::binary);
  if(!patterns)
  {
    std::cerr << "Error opening pattern file!" << std::endl;
    return 3;
  }
  PatternClassifier classifier(handler.write ? handler.patterns_name : "");

  std::vector<std::string> rows;
  readRows(patterns, rows, true);

  usint total = 0, n = rows.size();
  usint found = 0, forward = 0, reverse = 0;
  usint forward_occurrences = 0, reverse_occurrences = 0;

  double start = readTimer();
  for(usint i = 0; i < n; i++)
  {
    total += rows[i].length();
    bool match = false;

    if(handler.k == 0)
    {
      pair_type result = bwasearch.find(rows[i], false, handler.skip);
      if(bwasearch.handleOccurrences(result, handler.locate, forward_occurrences))
      {
        found++; forward++; match = true;
        if(handler.write) { classifier.forward(rows[i]); }
      }
      if(handler.reverse_complement)
      {
        pair_type reverse_result = bwasearch.find(rows[i], true, handler.skip);
        if(bwasearch.handleOccurrences(reverse_result, handler.locate, reverse_occurrences))
        {
          if(isEmpty(result)) { found++; }
          reverse++; match = true;
          if(handler.write) { classifier.reverse(rows[i]); }
        }
      }
      if(handler.write && !match) { classifier.notfound(rows[i]); }
    }
    else
    {
      std::vector<pair_type>* results;
      if(handler.penalties)
      {
        results = bwasearch.weightedFind(rows[i], handler.k, handler.indels);
      }
      else
      {
        results = bwasearch.find(rows[i], handler.k, handler.indels);
      }
      if(bwasearch.handleOccurrences(results, handler.locate, forward_occurrences))
      {
        found++;
      }
      delete results;
    }
  }
  double time = readTimer() - start;
  double megabases = total / (double)MILLION;

  std::cout << "Patterns:     " << n << " (" << (n / time) << " / sec)" << std::endl;

  std::cout << "Total size:   " << megabases << " megabases";
  if(!(handler.locate))
  {
    std::cout << " (" << (megabases / time) << " megabases / sec)";
  }
  std::cout << std::endl;

  std::cout << "Found:        " << found;
  if(handler.k == 0 && handler.reverse_complement)
  {
    std::cout << " (" << forward << " forward, " << reverse << " reverse complement)";
  }
  std::cout << std::endl;

  if(handler.locate)
  {
    std::cout << "Occurrences:  " << (forward_occurrences + reverse_occurrences) << " (";
    if(handler.k == 0 && handler.reverse_complement)
    {
      std::cout << forward_occurrences << " forward, " << reverse_occurrences << " reverse complement, ";
    }
    std::cout << ((forward_occurrences + reverse_occurrences) / time) << " / sec)" << std::endl;
  }

  std::cout << "Time:         " << time << " seconds" << std::endl;
  std::cout << std::endl;

  return 0;
}
